'''The files module contains functions related to handling audio data files, for example loading audio files, saving audio files, and examing and reformatting audio files.
'''
import numpy as np
import random
import collections
import math 
import pathlib
from scipy.io.wavfile import write, read
from scipy.signal import resample
import soundfile as sf
import librosa

import os, sys, tarfile
import inspect
currentdir = os.path.dirname(os.path.abspath(
    inspect.getfile(inspect.currentframe())))
packagedir = os.path.dirname(currentdir)
sys.path.insert(0, packagedir)

import soundpy as sp



def loadsound(filename, sr = None, mono = True, dur_sec = None, 
              remove_dc = True, use_scipy = False):
    '''Loads sound file with scipy.io.wavfile.read or librosa.load (default librosa)
    
    Parameters
    ----------
    filename : str
        The filename of the sound to be loaded
    
    sr : int, optional
        The desired sample rate of the audio samples. If None, 
        the sample rate of the audio file will be used.

    mono : bool
        If True, the samples will be loaded in mono sound. If False,
        if the samples are in stereo, they will be loaded in stereo sound.

    dur_sec : int, float, optional
        The length in seconds of the audio signal.

    remove_dc_bias : bool
        If True, the mean is subtracted from the signal. This has shown to be
        very helpful when working with audio data. (default True)

    use_scipy : bool 
        If False, librosa will be used to load the audiofile. If True, 
        scipy.io.wavfile and/or soundfile will be used. If the sound file 
        is not compatible with scipy.io.wavfile.read, this functions converts 
        the file to .wav format and/or changes the bit depth to be compatible. 
        (default False)
        
    Returns
    -------
    data : nd.array [size=(num_samples,) or (num_samples, num_channels)]
        The normalized (between 1 and -1) sample data returned 
        according to the specified settings.
    sr : int 
        The sample rate of the loaded samples.
        
    See Also
    --------
    soundpy.files.prep4scipywavfile
        Prepares audio file for scipy.io.wavfile.read.
        
    soundpy.files.convert_audiofile
        Converts audio file to .wav format.
    
    soundpy.files.newbitdepth
        Converts audio file to specified bitdepth.
        
    soundpy.dsp.resample_audio
        Resampe audio data to a specified sample rate.
        
    soundpy.files.list_possibleformats
        Lists the possible formats to load with soundpy.loadsound
        
    librosa.load
        The package used to load sound data by default. See `librosa`.
        
    scipy.io.wavfile.read
        The package used to load sound if `use_scipy` is set to True.
        See `scipy`.

    soundpy.dsp.remove_dc_bias
        Removes the 'direct current' bias from the signal.
        
    Todo
    ----
    Make librosa data and scipy.io.wavfile data more similar
        https://stackoverflow.com/questions/54482346/reading-a-wav-file-with-scipy-and-librosa-in-python
    '''
    if not use_scipy:
        # the sample data will be a litle different from scipy.io.wavfile
        # as librosa does a litle extra work with the data
        data, sr = librosa.load(filename, sr=sr, mono=mono, duration=dur_sec)
        if mono is False and len(data.shape) > 1: 
            if data.shape[0] < data.shape[1]:
                import warnings
                msg = '\nWARNING: Most functionality has not been tested '+\
                    'with stereo sound. Many functions may fail or not work as '+\
                        'expected. Apologies for the inconvenience!'
                warnings.warn(msg)
                # change shape from (channels, samples) to (samples, channels)
                data = data.T
        if remove_dc:
            data = sp.dsp.remove_dc_bias(data)
        return data, sr
    try:
        sr2, data = read(filename)
        if sr:
            if sr2 != sr:
                data, sr2 = sp.dsp.resample_audio(data, 
                                          sr_original = sr2, 
                                          sr_desired = sr)
                assert sr2 == sr
        else:
            sr = sr2
    except ValueError:
        print("Converting {} to wavfile".format(filename))
        try:
            filename = sp.files.convert_audiofile(filename, overwrite=False)
        except RuntimeError as e:
            raise RuntimeError('Try setting `use_scipy` to False in soundpy.loadsound().')
        try:
            data, sr = loadsound(filename, sr=sr, mono=mono, dur_sec=dur_sec)
            print("File saved as {}".format(filename))
        except ValueError:
            print("Ensure bitdepth is compatible with scipy library")
            filename = sp.files.newbitdepth(filename, overwrite=False)
            data, sr = loadsound(filename, sr=sr, mono=mono, dur_sec=dur_sec)
    
    # scipy loads data in shape (num_samples, num_channels)
    # don't need to transpose as for librosa
    if mono and len(data.shape) > 1:
        if data.shape[1] > 1:
            data = sp.dsp.stereo2mono(data)
    if not mono and len(data.shape) > 1:
            import warnings
            msg = '\nWARNING: Most functionality has not been tested '+\
                'with stereo sound. Many functions may fail or not work as '+\
                    'expected. Apologies for the inconvenience!'
            warnings.warn(msg)
    # scale samples to be between -1 and 1
    data = sp.dsp.scalesound(data, max_val= 1, min_val=-1)
    if dur_sec:
        numsamps = int(dur_sec * sr)
        data = sp.dsp.set_signal_length(data, numsamps)
    if remove_dc:
        data = sp.dsp.remove_dc_bias(data)
    return data, sr

def savesound(audiofile_name, signal_values, sr, remove_dc=True, 
              overwrite=False, use_scipy=False, **kwargs):
    """saves the wave at designated path

    Parameters
    ----------
    audiofile_name : str or pathlib.PosixPath
        path and name the audio is to be saved under. (.wav format)
    
    signal_values : ndarray
        values of real signal to be saved
    
    sr : int 
        sample rate of the audio samples.
    
    remove_dc : bool 
        If True, the mean is subtracted from the signal. (default True)
    
    overwrite : bool
        If True, audio with the same naem will be overwritten. (default False)
        
    use_scipy : bool
        If True, scipy.io.wavfile.write will be used. However, file conversion is 
        limited. Can only save .wav files. Otherwise soundfile.write will be used, 
        which can save audio under more audio fomats.
    
    **kwargs : additional keyword arguments
        The keyword arguments for soundfile.write:
        https://pysoundfile.readthedocs.io/en/latest/index.html?highlight=write#soundfile.write

    Returns
    -------
    audiofile_name : pathlib.PosixPath
        The new audiofile name
        
    See Also
    --------
    scipy.io.wavfile.write
    
    soundpy.files.conversion_formats
        Lists the possible formats to save audio files if `use_scipy` is False.
        
    soundpy.dsp.remove_dc_bias
        Removes the 'direct current' bias from the signal.
    """
    audiofile_name = sp.utils.string2pathlib(audiofile_name)
    if os.path.exists(audiofile_name) and overwrite is False:
        raise FileExistsError('Filename {} already exists.'.format(audiofile_name)+\
            '\nSet `overwrite` to True in function savesound() to overwrite.')
    directory = audiofile_name.parent
    directory = sp.utils.check_dir(directory, make=True)
    if remove_dc:
        signal_values = sp.dsp.remove_dc_bias(signal_values)
    if use_scipy:
        write(audiofile_name, sr, signal_values)
    else: 
        # check to see if filename extension and format match:
        # if not, warning
        if 'format' in kwargs:
            if kwargs['format'].lower() != audiofile_name.suffix[1:]:
                audiofile_name_orig = audiofile_name
                audiofile_name = replace_ext(audiofile_name, kwargs['format'].lower())
                import warnings
                message = 'The desired format does not match the new file name: '+\
                    '\nDesired format: {}'.format(kwargs['format'])+\
                        '\nFilename: {}'.format(audiofile_name_orig) +\
                            '\nExtension will be adjusted to the format setting: '+\
                                '\n{}'.format(audiofile_name)
                warnings.warn(message)
        sf.write(audiofile_name, signal_values, sr, **kwargs)
    return audiofile_name

def get_file_format(audiofile):
    '''Use soundfile to get file format.
    '''
    so = sf.SoundFile(audiofile)
    return so.format

def list_possibleformats(use_scipy=False):
    if not use_scipy:
        return(['.wav', '.aiff', '.flac', '.ogg','.m4a','.mp3'])
    else:
        return(['.wav', '.aiff', '.flac', '.ogg'])

def list_audioformats():
    msg = '\nsoundpy can work with the following file types: '+\
        ', '.join(get_compatible_formats())+ \
            '\nSo far, functionality does not work with the following types: '+\
                ', '.join(get_incompatible_formats())
    return msg

# TODO finish
def audiofiles_present(directory, recursive=False):
    '''Checks to see if audio files are present. 
    
    Parameters
    ----------
    directory : str or pathlib.PosixPath
        The directory to look for audio.
        
    recursive : bool
        If True, all nested directories will be checked as well. (default False)
        
    Returns
    -------
    bool 
        True if audio is present; otherwise False.
    '''
    directory = sp.utils.string2pathlib(directory)

def collect_audiofiles(directory, hidden_files = False, wav_only=False, recursive=False):
    '''Collects all files within a given directory.
    
    This includes the option to include hidden_files in the collection.
    
    Parameters
    ----------
    directory : str or pathlib.PosixPath
        The path to where desired files are located.
    hidden_files : bool 
        If True, hidden files will be included. If False, they won't.
        (default False)
    wav_only : bool 
        If True, only .wav files will be included. Otherwise, no limit
        on file type. 
    
    Returns
    -------
    paths_list : list of pathlib.PosixPath objects
        Sorted list of file pathways.
    '''
    if not isinstance(directory, pathlib.PosixPath):
        directory = pathlib.Path(directory)
    paths_list = []
    # allow all data types to be collected (not only .wav)
    if wav_only:
        filetype = '*.wav'
    else: 
        filetype = '*'
    if recursive:
        filetype = '**/' + filetype
    for item in directory.glob(filetype):
        paths_list.append(item)
    # pathlib.glob collects hidden files as well - remove them if they are there:
    if not hidden_files:
        paths_list = [x for x in paths_list if x.stem[0] != '.']
    # ensure only audiofiles:
    paths_list = sp.files.ensure_only_audiofiles(paths_list)
    return paths_list


def collect_zipfiles(directory, hidden_files = False, ext='tgz', recursive=False):
    '''Collects all zipfiles within a given directory.
    
    This includes the option to include hidden_files in the collection.
    
    Parameters
    ----------
    directory : str or pathlib.PosixPath
        The path to where desired files are located.
    hidden_files : bool 
        If True, hidden files will be included. If False, they won't.
        (default False)
    wav_only : bool 
        If True, only .wav files will be included. Otherwise, no limit
        on file type. 
    
    Returns
    -------
    paths_list : list of pathlib.PosixPath objects
        Sorted list of file pathways.
    '''
    if not isinstance(directory, pathlib.PosixPath):
        directory = pathlib.Path(directory)
    paths_list = []
    # allow all data types to be collected (not only .wav)
    if ext[0] == '.':
        filetype = '*' + ext
    else: 
        filetype = '*.' + ext
    if recursive:
        filetype = '**/' + filetype
    for item in directory.glob(filetype):
        paths_list.append(item)
    # pathlib.glob collects hidden files as well - remove them if they are there:
    if not hidden_files:
        paths_list = [x for x in paths_list if x.stem[0] != '.']
    return paths_list
    
def ensure_only_audiofiles(audiolist):
    possible_extensions = sp.files.list_possibleformats(use_scipy=False)
    audiolist_checked = [x for x in audiolist if pathlib.Path(x).suffix in possible_extensions]
    if len(audiolist_checked) < len(audiolist):
        import warnings
        message = 'Some files did not match those acceptable by this program. '+\
            '(i.e. non-audio files) The number of files not included: '+\
                '{}'.format(len(audiolist)-len(audiolist_checked))
        warnings.warn(message)
    return audiolist_checked

def prep4scipywavfile(filename, overwrite=False):
    '''Takes soundfile and saves it in a format compatible with scipy.io.wavfile
    
    Parameters
    ----------
    filename : str
        Filename of the soundfile to load with scipy.io.wavfile
    
    Returns
    -------
    filename : str
        Filename of the soundfile compatible with scipy.io.wavfile
    '''
    try:
        sr, data = read(filename)
        return filename
    except ValueError as e:
        import pathlib
        if pathlib.Path(filename).suffix.lower() != '.wav': 
            print("Converting file to .wav")
            filename = convert_audiofile(filename, overwrite=overwrite)
            print("Saved file as {}".format(filename))
        elif 'bitdepth' not in str(filename):
            print("Ensuring bitdepth is compatible with scipy library")
            filename = newbitdepth(filename, overwrite=overwrite)
            print("Saved file as {}".format(filename))
        else:
            #some other error
            raise e
        filename = prep4scipywavfile(filename, overwrite=overwrite)
    return filename

def conversion_formats():
    '''Lists the formats available for conversion. 
    
    soundpy uses soundfile to convert files; therefore, whatever
    available formats soundfile has will be listed here.
    
    Examples
    --------
    >>> conversion_formats
    {'AIFF': 'AIFF (Apple/SGI)',
    'AU': 'AU (Sun/NeXT)',
    'AVR': 'AVR (Audio Visual Research)',
    'CAF': 'CAF (Apple Core Audio File)',
    'FLAC': 'FLAC (Free Lossless Audio Codec)',
    'HTK': 'HTK (HMM Tool Kit)',
    'SVX': 'IFF (Amiga IFF/SVX8/SV16)',
    'MAT4': 'MAT4 (GNU Octave 2.0 / Matlab 4.2)',
    'MAT5': 'MAT5 (GNU Octave 2.1 / Matlab 5.0)',
    'MPC2K': 'MPC (Akai MPC 2k)',
    'OGG': 'OGG (OGG Container format)',
    'PAF': 'PAF (Ensoniq PARIS)',
    'PVF': 'PVF (Portable Voice Format)',
    'RAW': 'RAW (header-less)',
    'RF64': 'RF64 (RIFF 64)',
    'SD2': 'SD2 (Sound Designer II)',
    'SDS': 'SDS (Midi Sample Dump Standard)',
    'IRCAM': 'SF (Berkeley/IRCAM/CARL)',
    'VOC': 'VOC (Creative Labs)',
    'W64': 'W64 (SoundFoundry WAVE 64)',
    'WAV': 'WAV (Microsoft)',
    'NIST': 'WAV (NIST Sphere)',
    'WAVEX': 'WAVEX (Microsoft)',
    'WVE': 'WVE (Psion Series 3)',
    'XI': 'XI (FastTracker 2)'}
    '''
    return sf.available_formats()

def convert_audiofile(filename, format_type=None, sr=None, new_dir=False, overwrite=False, use_scipy=False, **kwargs):
    '''Converts and saves soundfile as .wav type in same or new directory.
    
    Parameters
    ----------
    filename : str or pathlib.PosixPath
        The filename of the audiofile to be converted to .wav type
        
    format_type : str 
        The format to convert the audio file to. See soundpy.files.conversion_formats. 
        (defaults to 'wav')
    
    new_dir : str, pathlib.PosixPath, optional 
        If False, the converted files will be saved in same directory as originals.
        If a path is provided, the converted files will be saved there. If no such directory
        exists, one will be created.
    
    sr : int, optional
        The sample rate to be applied to the signal. If none supplied, the sample rate 
        of the original file will be used.
        
    **kwargs : additional keyword arguments
        The keyword arguments for soundfile.write:
        https://pysoundfile.readthedocs.io/en/latest/index.html?highlight=write#soundfile.write
    
        
    Returns 
    -------
    f_wavfile : pathlib.PosixPath
        The filename / path where the audio file is saved.
        
    Examples
    --------
    >>> audiofile = './example/audio.wav'
    # in same directory
    >>> audiofile_flac = sp.files.convert_audiofile(audiofile, format_type='flac')
    >>> audiofile_flac
    PosixPath('example/audio.flac')
    # in new directory
    >>> audiofile_flac = sp.files.convert_audiofile(audiofile, format_type='flac', 
                                                     new_dir = './examples2/')
    >>> audiofile_flac
    PosixPath('examples2/audio.flac')
    >>> # can establish desired conversion format in `new_dir`
    >>> audiofile_ogg = sp.files.convert_audiofile(audiofile,
                                                     new_dir = './examples2/audio.ogg')
    >>> audiofile_ogg
    PosixPath('audiodata2/audio.ogg')
    
    See Also
    --------
    soundpy.files.conversion_formats
        Lists the possible formats to convert audio files.
        
    soundpy.files.list_possibleformats
        Lists the possible formats to load with soundpy.loadsound
    '''
    import pathlib
    import os
    try:
        f = pathlib.Path(filename)
    except TypeError:
        raise TypeError('Function convert_audiofile expected input of type string '+\
            'or a pathlib object, not type {}.'.format(type(filename)))
    if not f.suffix:
        raise TypeError('Function convert_audiofile expected a path with an '+\
            'audio extension, not input: \n', filename)
    if not f.suffix in sp.files.list_possibleformats(use_scipy=False):
        raise TypeError('This software cannot process audio in {}'.format(f.suffix)+\
            ' format. We apologize for the inconvenience.')
    # ensure filename exists:
    if not os.path.exists(filename):
        raise FileNotFoundError('Could not find audio file at the following '+\
            'location\n{}'.format(filename))
    
    # set format_type to match suffix of new pathway (if it is not set already)
    if new_dir:
        new_dir = pathlib.Path(new_dir)
        if new_dir.suffix:
            if format_type is None:
                format_type = new_dir.suffix[1:].upper()
    
    # set default if format_type is None
    if format_type is None:
        format_type = 'WAV'
    
    # establish the path to save updated file.
    if new_dir:
        # check if new_dir is a directory or filename
        if new_dir.suffix:
            new_filename = new_dir
            new_dir = new_dir.parent
            # check for mismatch in file type and file name
            if format_type.lower() != new_filename.suffix[1:].lower():
                import warnings
                message = '\nWARNING: `format_type` {} '.format(format_type) +\
                    'does not match format of `new_dir`: \n{}.\n'.format(new_filename) + \
                        'Saving according to the `format_type`.'
                warnings.warn(message)
                new_filename = replace_ext(new_filename, format_type.lower())
        else:
            new_f = replace_ext(f, format_type.lower())
            # change filename to new directory
            new_filename = new_dir.joinpath(new_f.name)
        # check to make sure new_dir exists
        new_dir = sp.utils.check_dir(new_dir, make=True)
    else:
        new_filename = replace_ext(f, format_type.lower())
        
    # load audio samples with soundfile, then save them as new audio file.
    try:
        data, sr = sf.read(filename, samplerate=sr)
    except RuntimeError as e:
        data, sr = librosa.load(filename, sr=sr)

    if not overwrite:
        if os.path.exists(new_filename):
            import warnings
            message = '\nWARNING: File {} already exists at this '.format(new_filename)+\
                'location. Set `overwrite` to True to overwrite it. Not overwritten.'
            warnings.warn(message)
            return new_filename
    if not use_scipy:
        format = format_type.upper()
        
    new_filename = sp.savesound(new_filename, data, sr, 
                                  overwrite=overwrite, use_scipy=use_scipy,
                                  format=format)
    return new_filename

def replace_ext(filename, extension):
    '''Adds or replaces an extension in the filename
    
    Parameters
    ----------
    filename : str or pathlib.PosixPath
        Filename with the missing or incorrect extension
    extension : str
        The correct extension for the given filename.
    
    Returns
    -------
    file_newext : pathlib.PosixPath
        The filename with the new extension
    '''
    if isinstance(filename, str):
        import pathlib
        filename = pathlib.Path(filename)
    #filestring = str(filename)[:len(str(filename))-len(filename.suffix)]
    f_dir = filename.parent
    f_name = filename.stem
    if extension[0] != '.':
        extension = '.'+extension
    new_filename = f_dir.joinpath(f_name + extension)
    return new_filename

def match_ext(filename1, filename2):
    '''Matches the file extensions. 
    
    If both have extensions, default set to that of `filename1`.
    '''
    import pathlib
    f1 = pathlib.Path(filename1)
    f2 = pathlib.Path(filename2)
    if not f1.suffix:
        if not f2.suffix:
            raise TypeError('No file extension provided. Check the filenames.')
        else: 
            extension = f2.suffix
    else: 
        extension = f1.suffix 
    if f1.suffix != extension:
        f1 = replace_ext(f1, extension)
    else:
        f1 = str(f1)
    if f2.suffix != extension:
        f2 = replace_ext(f2, extension)
    else:
        f2 = str(f2)
    return f1, f2

def newbitdepth(wave, bitdepth=16, newname=None, overwrite=False):
    '''Convert bitdepth to 16 or 32, to ensure compatibility with scipy.io.wavfile
    
    Scipy.io.wavfile is easily used online, for example in Jupyter notebooks.
    
    Reference
    ---------
    https://stackoverflow.com/questions/44812553/how-to-convert-a-24-bit-wav-file-to-16-or-32-bit-files-in-python3
    '''
    if bitdepth == 16:
        newbit = 'PCM_16'
    elif bitdepth == 32:
        newbit = 'PCM_32'
    else:
        raise ValueError('Provided bitdepth is not an option. Available bit depths: 16, 32')
    data, sr = sf.read(wave)

    if overwrite:
        if newname is None:
            newname = wave
        sf.write(newname, data, sr, subtype=newbit)
        savedname = newname
    else:
        if not newname:
            newname = adjustname(wave, adjustment='_bitdepth{}'.format(bitdepth))
            print("No new filename provided. Saved file as '{}'".format(newname))
            sf.write(newname, data, sr, subtype=newbit)
        elif newname:
            #make sure new extension matches original extension
            wave, newname = match_ext(wave, newname)
            sf.write(newname, data, sr, subtype=newbit)
        savedname = newname
    return savedname

def adjustname(filename, adjustment=None):
    '''Adjusts filename.
    
    Parameters
    ----------
    filename : str
        The filename to be adjusted
    adjustment : str, optional
        The adjustment to add to the filename. If None, 
        the string '_adj' will be added.
    
    Returns
    -------
    fname : str 
        The adjusted filename with the original extension
        
    Examples
    --------
    >>> adjustname('happy.md')
    'happy_adj.md'
    >>> adjustname('happy.md', '_not_sad')
    'happy_not_sad.md'
    '''
    import pathlib
    f = pathlib.Path(filename)
    fname = f.stem
    if adjustment:
        fname += adjustment
    else:
        fname += '_adj'
    fname += f.suffix
    return fname

def extract(tar_url, extract_path='.'):
    tar = tarfile.open(tar_url, 'r')
    for item in tar:
        tar.extract(item, extract_path)
        if item.name.find(".tgz") != -1 or item.name.find(".tar") != -1:
            extract(item.name, "./" + item.name[:item.name.rfind('/')])

def delete_dir_contents(directory, remove_dir = False):
    '''
    https://stackoverflow.com/a/28834214
    '''
    d = sp.utils.string2pathlib(directory)
    for sub in d.iterdir():
        if sub.is_dir():
            delete_dir_contents(sub)
        else:
            sub.unlink()
    if remove_dir:
        d.rmdir()
        
def matching_filenames(list1, list_of_lists):
    list1_files = []
    if isinstance(list1[0], tuple):
        for item in list1:
            if len(item) != 2:
                # ensures expected length of 2: (encoded_label, pathway)
                raise ValueError('Expected a list of tuple pairs: encoded '+\
                    'label and associated pathway. Received tuple of length ', len(item))
            # checks to ensure encoded label comes first
            if isinstance(item[0], int) or isinstance(item[0], str) and item[0].isdigit():
                list1_files.append(item[1])
    elif isinstance(list1[0], str) or isinstance(list1[0], pathlib.PosixPath) or isinstance(list1[0], pathlib.PurePath):
        list1_files = list1
    other_lists_files = []
    for l in list_of_lists:
        if isinstance(l[0], tuple):
            for item in l:
                if len(item) != 2:
                    # ensures expected length of 2: (encoded_label, pathway)
                    raise ValueError('Expected a list of tuple pairs: encoded '+\
                        'label and associated pathway. Received tuple of length ', len(item))
                # checks to ensure encoded label comes first
                if isinstance(item[0], int) or isinstance(item[0], str) and item[0].isdigit():
                    # ensure pathway is string, not pathlib (for iteration purposes)
                    other_lists_files.append(str(item[1]))
        elif isinstance(l[0], str) or isinstance(l[0], pathlib.PosixPath) or isinstance(l[0], pathlib.PurePath):
            other_lists_files.append(l)
    # ensure list of lists is flat list
    flatten = lambda l: [item for sublist in l for item in sublist]
    other_lists_files = flatten(other_lists_files)
    contanimated_files = []
    for item in list1_files:
        if isinstance(item, str):
            item = sp.string2pathlib(item)
        fname = item.stem
        fname_parts = fname.split('-')
        fname_head = fname_parts[:-1]
        fname_head = '-'.join(fname_head)
        for i in other_lists_files:
            if fname_head in i:
                contanimated_files.append(item)
                break
    return contanimated_files
        
def remove_contaminated_files(list1, contaminated_files):
    # ensure files are strings not pathlib.PosixPath objects
    contaminated_files = [str(x) for x in contaminated_files]
    remove_idx = []
    if isinstance(list1[0], tuple):
        for i, tuple_pair in enumerate(list1):
            pathway = str(tuple_pair[1])
            if pathway in contaminated_files:
                remove_idx.append(i)
    list_uncont = [x for j, x in enumerate(list1) if j not in remove_idx]
    return list_uncont



import tensorflow
from tensorflow.keras.models import Model
from tensorflow.keras.models import load_model 
import numpy as np
import matplotlib.pyplot as plt

import os, sys
import inspect
currentdir = os.path.dirname(os.path.abspath(
    inspect.getfile(inspect.currentframe())))
packagedir = os.path.dirname(currentdir)
sys.path.insert(0, packagedir)
import soundpy as pyst


def featuremaps(features, model, image_dir='./feature_maps/'):
    '''Saves the feature maps of each convolutional layer as .png file.
    
    References
    ----------
    Brownlee, Jason (2019, May, 6). How to Visualize Filters and Feature
    Maps in Convolutional Neural Networks. Machine Learning Mastery.
    https://machinelearningmastery.com/how-to-visualize-filters-and-feature-maps-in-convolutional-neural-networks/
    '''
    conv_idx = []
    for i in range(len(model.layers)):
        layer = model.layers[i]
        if 'conv' in layer.name:
            conv_idx.append(i)
    for idx in conv_idx:
        model_featmaps = Model(inputs = model.inputs,
                    outputs = model.layers[idx].output)
        featuremaps = model_featmaps.predict(features)
        for i in range(featuremaps.shape[-1]):
            plt.clf()
            plt.imshow(featuremaps[0,:,:,i], cmap='gray')
            image_dir = sp.utils.check_dir(image_dir, make=True)
            image_path = image_dir.joinpath('layer_{}'.format(idx),
                                            'featmap_{}.png'.format(i))
            image_par = sp.utils.check_dir(image_path.parent, make=True)
            plt.savefig(image_path)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

dependencies=''
with open("requirements.txt","r") as f:
        dependencies = f.read().splitlines()

# This call to setup() does all the work
setup(
    name="soundpy",
    version="0.1.0a3",
    description="A research-based framework for exploring sound as well as machine learning in the context of sound.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/a-n-rose/Python-Sound-Tool",
    author="Aislyn Rose",
    author_email="rose.aislyn.noelle@gmail.com",
    license="AGPL-3.0",
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=("tests","docs", "jupyter_notebooks")),
    include_package_data=True,
    install_requires=dependencies,
    python_requires=">=3.6.9",
)



'''The models.modelsetup module containes functionality for preparing for training a model
'''

import os, sys
import inspect
currentdir = os.path.dirname(os.path.abspath(
    inspect.getfile(inspect.currentframe())))
packagedir = os.path.dirname(currentdir)
sys.path.insert(0, packagedir)
import tensorflow
from tensorflow.keras.callbacks import EarlyStopping, CSVLogger,\
    ModelCheckpoint, TensorBoard
import numpy as np
import soundpy as sp

###############################################################################

def setup_layers(num_features, num_layers, kernel_shape = (3,3), max_feature_map = 64):
    '''Sets up `feature_maps` and `kernels` for 1 or more layered convolutional neural networks.
    
    Parameters
    ----------
    num_features : int 
        The number of features used to train the model. This will be used to set the number 
        of `feature_maps` for each layer. 
        
    num_layers : int 
        The number of layers desired 
        
    kernel_shape : tuple or int 
        The shape of the desired kernel 
        
    max_feature_map : int 
        The maximum size of feature map / filter. This depends on the system and is relevant for
        processing higher definition features, such as STFT features. If this is set too large 
        given memory restraints, training may be 'killed'.
        
    Returns
    -------
    feature_maps : list 
        List of feature maps or filters that will be applied to each layer of the network.
        
    kernels : list 
        List of kernels that will be applied to each layer of the network. Matches length of 
        `feature_maps`
        
    Warning
    -------
    If `num_features` is larger than the `max_feature_map`. The `num_features` is usually used to
    set the first feature map, but if too large, will be reduced to be lower than `max_feature_map`.
        
    '''
    if num_features > max_feature_map:
        num_features_orig = num_features
        while num_features > max_feature_map:
            num_features //= 1.1
        import warnings
        msg = '\nWARNING: feature maps will not be set according to `num_features` '+\
            'as `num_features` is too high ({})'.format(num_features_orig)+\
                ' The first feature map will be reduced to {}'.format(num_features)
        warnings.warn(msg)
    feature_maps = []
    kernels = []
    for i in range(num_layers):
        if i == 0:
            feature_maps.append(num_features)
            kernels.append(kernel_shape)
        else:
            feature_maps.append(feature_maps[i-1]//2)
            kernels.append(kernel_shape)
    return feature_maps, kernels

def setup_callbacks(early_stop=True, patience=15, log=True,
                        log_filename=None, append=True, save_bestmodel=True, 
                        best_modelname=None,monitor='val_loss', verbose=1, save_best_only=True, mode='min', tensorboard=True,
                        write_images=False, x_test=None, y_test=None,
                        batch_size = None, embedded_layer_name = None):
    '''Easy set up of early stopping, model logging, and saving best model.
    
    Parameters
    ----------
    early_stop : bool 
        Whether or not the model should stop if training is not improving 
        (default True)
    patience : int 
        The number of epochs the model should complete without improvement
        before stopping training. (default 15)
    log : bool 
        If true, the accuracy, loss, and (if possible) the val_accuracy and 
        val_loss for each epoch will be saved in a .csv file. (default True)
    log_filename : str or pathlib.PosixPath, optional
        The filename the logging information will be stored. If None, 
        the date will be used as a unique .csv filename in a subfolder 
        'model_logs' in the local directory.
    save_bestmodel : bool
        If True, the best performing model will be saved.
    best_modelname : str or pathlib.PosixPath
        The name to save the best model version under. If None, the date
        will be used to create a unique .h5 filename and it will be saved
        in a subfolder 'best_models' in the local directory.
    monitor : str 
        The metric to be used to measure model performance. 
        (default 'val_loss'
    verbose : bool
        If True, the state of the model will be printed. (default True)
    save_best_only : bool 
        If True, the best performing model will overwrite any previously 
        saved 'best model'.
    mode : str 
        If `monitor` is set to 'val_loss', this should be set to 'min'.
        If `monitor``is set to 'val_acc', this should be set to 'max'. If 
        `mode` is set to 'auto', the direction will be inferred. 
        (default 'min')
    tensorboard : bool 
        If True, logs for TensorBoard will be made.
        
    Returns
    -------
    callbacks : # TODO what data type is this?
        The callbacks ready to be applied to Keras model training.
     '''
    callbacks = []
    if early_stop:
        early_stopping_callback = EarlyStopping(
            monitor=monitor,
            patience=patience)
        callbacks.append(early_stopping_callback)
    if log or log_filename:
        if log_filename is None:
            # create directory to store log files:
            log_dir = sp.utils.check_dir('./model_logs/', 
                                           make=True, 
                                           write_into=True)
            log_filename = '{}log_'.format(log_dir)+\
                sp.utils.get_date()+'.csv'
        # TODO test if pathlib.PosixPath won't work
        # if pathlib object, turn into string
        # TODO do some sort of check to ensure path is valid?
        if not isinstance(log_filename, str):
            log_filename = str(log_filename)
        csv_logging = CSVLogger(filename=log_filename, append=append)
        callbacks.append(csv_logging)
    if save_bestmodel:
        if best_modelname is None:
            model_dir = sp.utils.check_dir('./best_models/', 
                                             make=True, 
                                             write_into=True)
            best_modelname = '{}model_'.format(model_dir)+\
                sp.utils.get_date()+'.h5'
        # TODO test if pathlib.PosixPath won't work
        # if pathlib object, turn into string
        # TODO do some sort of check to ensure path is valid?
        if not isinstance(best_modelname, str):
            best_modelname = str(best_modelname)
        checkpoint_callback = ModelCheckpoint(best_modelname, monitor=monitor,
                                                verbose=verbose,
                                                save_best_only=save_best_only,
                                                mode=mode)
        callbacks.append(checkpoint_callback)
    if tensorboard:
        if log_filename is not None:
            log_filename = sp.utils.string2pathlib(log_filename)
            log_dir = log_filename.parent
            log_dir = log_dir.joinpath('tb_logs/')
        else:
            log_dir = sp.utils.string2pathlib('./model_logs/tb_logs/')
        log_dir = sp.utils.check_dir(log_dir, make=True)
        if x_test is not None and y_test is not None:
            with open(os.path.join(log_dir, 'metadata.tsv'), 'w') as f:
                np.savetxt(f, y_test)
            embeddings_freq = 1
            if embedded_layer_name is None:
                embedded_layer_name = 'features'
            embeddings_layer_names = [embedded_layer_name]
            embeddings_metadata = 'metadata.tsv'
            embeddings_data = x_test
        else:
            embeddings_freq = 0
            embeddings_layer_names = None
            embeddings_metadata = None
            embeddings_data = None
        tb = TensorBoard(log_dir, batch_size = batch_size, write_images=write_images, 
                         embeddings_freq = embeddings_freq, 
                         embeddings_layer_names = embeddings_layer_names, 
                         embeddings_metadata = embeddings_metadata,
                         embeddings_data = embeddings_data)
        callbacks.append(tb)
    if callbacks:
        return callbacks
    return None



'''The models.template_models module contains functions for building (ideally research-based) models.
'''

import os, sys
import inspect
import pathlib
import numpy as np
# for building and training models
import tensorflow
from tensorflow.keras import applications
from tensorflow.keras.models import Sequential, load_model, Model
from tensorflow.keras.layers import Dense, Conv2D, Flatten, Dropout, Conv2DTranspose, \
    LSTM, MaxPooling2D, TimeDistributed
from tensorflow.keras.constraints import max_norm
 
currentdir = os.path.dirname(os.path.abspath(
    inspect.getfile(inspect.currentframe())))
packagedir = os.path.dirname(currentdir)
sys.path.insert(0, packagedir)

import soundpy as pyst

###############################################################################

            
def adjust_layers_cnn(**kwargs):
    '''Reduces layers of CNN until the model can be built.

    If the number of filters for 'mfcc' or 'fbank' is in the lower range
    (i.e. 13 or so), this causes issues with the default settings of
    the cnn architecture. The architecture was built with at least 40
    filters being applied during feature extraction. To deal with this
    problem, the number of CNN layers are reduced. 
    
    Parameters
    ----------
    **kwargs : Keyword arguments 
        Keyword arguments for soundpy.models.template_models.cnn_classifier
        
    Returns
    -------
    settings : dict 
        Updated dictionary with relevant settings for model.
        
    References
    ----------
    https://github.com/pgys/NoIze
    '''
    
    settings = dict(kwargs)
    try: 
        assert len(settings['feature_maps']) == len(settings['kernel_size'])
    except AssertionError:
        raise ValueError('Expect `feature_maps` and `kernel_size` to have same length.')
    if len(settings['feature_maps']) > 1:
        settings['feature_maps'] = kwargs['feature_maps'][:-1]
    else:
        raise ValueError('CNN Model cannot be trained with set number of '+\
            'features and feature maps.')
    if len(settings['kernel_size']) > 1:
        settings['kernel_size'] = kwargs['kernel_size'][:-1]
    return settings

def cnn_classifier(feature_maps = [40, 20, 10], 
                   kernel_size =  [(3, 3), (3, 3), (3, 3)],
                   strides = 2,
                   activation_layer = 'relu', 
                 activation_output = 'softmax', 
                 input_shape = (79, 40, 1), 
                 num_labels = 3, 
                 dense_hidden_units=100, dropout=0.25):
    '''Build a single or multilayer convolutional neural network.
    
    Parameters
    ----------
    feature_maps : int or list 
        The filter or feature map applied to the data. One feature map per
        convolutional neural layer required. For example, a list of length
        3 will result in a three-layer convolutional neural network.
    kernel_size : tuple or list of tuples
        Must match the number of feature_maps. The size of each corresponding feature map.
    strides : int
    
    activation_layer : str 
        (default 'relu')
    activation_outpu : str 
        (default 'softmax')
    input_shape : tuple
        The shape of the input 
    dense_hidden_units : int, optional
    
    dropout : float, optional
        Reduces overfitting
        
    Returns
    -------
    model : tf.keras.Model
        Model ready to be compiled.
        
    settings : dict 
        Dictionary with relevant settings for model.
        
    Warning
    -------
    If number features are not compatible with number of layers, warning raised and 
    layers adjusted. E.g. For lower number of MFCC features this will likely be applied if number
    of layers is greater than 1.
        
    References
    ----------
    A. Sehgal and N. Kehtarnavaz, "A Convolutional Neural Network 
    Smartphone App for Real-Time Voice Activity Detection," in IEEE Access, 
    vol. 6, pp. 9017-9026, 2018. 
    '''
    try:
        settings = dict(feature_maps = feature_maps, 
                        kernel_size = kernel_size,
                        strides = strides,
                        activation_layer = activation_layer,
                        activation_output = activation_output,
                        input_shape = input_shape,
                        num_labels = num_labels,
                        dense_hidden_units = dense_hidden_units,
                        dropout = dropout)
        model = Sequential()
        if not isinstance(feature_maps, list):
            feature_maps = list(feature_maps)
        if not isinstance(kernel_size, list):
            kernel_size = list(kernel_size)
        assert len(feature_maps) == len(kernel_size)
        for i, featmap in enumerate(feature_maps):
            if i == 0:
                model.add(Conv2D(featmap,
                        kernel_size[i],
                        strides = strides,
                        activation = activation_layer,
                        input_shape = input_shape))
            else:
                model.add(Conv2D(featmap,
                        kernel_size[i],
                        strides = strides,
                        activation = activation_layer))
        if dense_hidden_units is not None:
            model.add(Dense(dense_hidden_units))
        if dropout is not None:
            model.add(Dropout(dropout))
        model.add(Flatten())
        model.add(Dense(num_labels, activation = activation_output))   
    except ValueError:
        import warnings
        msg = '\nWARNING: number of layers ({}) incompatible with number'.format(len(feature_maps))+\
            ' of features or filters ({}). Reducing number of layers until'.format(feature_maps[0])+\
                ' model and number features / filters is compatible.'
        warnings.warn(msg)
        num_layers_orig = len(feature_maps)
        try:
            updated_settings = adjust_layers_cnn(**settings)
        except ValueError as e:
            print(e)
            return None, settings
        model, settings = cnn_classifier(**updated_settings)
        print('Updated number of layers: {}'.format(len(settings['feature_maps'])))
        
    return model, settings

# TODO: update model based on research
def autoencoder_denoise(input_shape,
                        kernel_size = (3,3),
                        max_norm_value = 2.0,
                        activation_function_layer = 'relu',
                        activation_function_output = 'sigmoid',
                        padding = 'same',
                        kernel_initializer = 'he_uniform'):
    '''Build a simple autoencoder denoiser.
    
    Parameters
    ----------
    input_shape : tuple
        Shape of the input data.
    max_norm_value : int or float
        
    Returns
    -------
    autoencoder : tf.keras.Model
        Model ready to be compiled
        
    References
    ----------
    Versloot, Christian (2019, December 19). Creating a Signal Noise Removal 
    Autoencoder with Keras. MachineCurve. https://www.machinecurve.com
    '''
    autoencoder = Sequential()
    autoencoder.add(Conv2D(128, kernel_size = kernel_size,
                            kernel_constraint = max_norm(max_norm_value),
                            activation = activation_function_layer,
                            kernel_initializer = kernel_initializer,
                            input_shape=input_shape))
    autoencoder.add(Conv2D(32, kernel_size = kernel_size,
                            kernel_constraint = max_norm(max_norm_value),
                            activation = activation_function_layer,
                            kernel_initializer = kernel_initializer))
    autoencoder.add(Conv2DTranspose(32, kernel_size = kernel_size,
                            kernel_constraint=max_norm(max_norm_value),
                            activation = activation_function_layer,
                            kernel_initializer = kernel_initializer))
    autoencoder.add(Conv2DTranspose(128, kernel_size = kernel_size,
                            kernel_constraint = max_norm(max_norm_value),
                            activation = activation_function_layer,
                            kernel_initializer = kernel_initializer))
    autoencoder.add(Conv2D(1, kernel_size = kernel_size,
                            kernel_constraint = max_norm(max_norm_value),
                            activation = activation_function_output, 
                            padding = padding))
    settings = dict(input_shape = input_shape,
                    kernel_size = kernel_size,
                    max_norm_value = max_norm_value,
                    activation_function_layer = activation_function_layer,
                    activation_function_output = activation_function_output,
                    padding = padding,
                    kernel_initializer = kernel_initializer)
    return autoencoder, settings

def resnet50_classifier(input_shape, num_labels, activation = 'softmax',
                        final_layer_name = 'features'):
    '''Simple image classifier built ontop of a pretrained ResNet50 model.
    
    References
    ----------
    Revay, S. & Teschke, M. (2019). Multiclass Language Identification using Deep 
    Learning on Spectral Images of Audio Signals. arXiv:1905.04348 [cs.SD]
    '''
    if len(input_shape) != 3:
        raise ValueError('ResNet50 expects 3D feature data, not {}D.'.format(
            len(input_shape)))
    if input_shape[-1] != 3:
        raise ValueError('ResNet50 expects 3 channels for RGB values, not {}.'.format(
            input_shape[-1]))
    base_model = applications.resnet50.ResNet50(weights=None, include_top=False,
                                                input_shape = input_shape)
    x = base_model.output 
    x = Flatten()(x)
    predictions = Dense(num_labels,
                        activation = activation,
                        name = final_layer_name)(x) # add name to this layer for TensorBoard visuals
    model = Model(inputs = base_model.input,
                  outputs = predictions)
    settings = dict(input_shape = input_shape,
                    num_labels = num_labels,
                    activation = activation,
                    final_layer_name = final_layer_name)
    return model, settings

def cnnlstm_classifier(num_labels, input_shape, lstm_cells, 
                       feature_map_filters = 32, kernel_size = (8,4),
                       pool_size = (3,3), dense_hidden_units = 60, 
                       activation_layer = 'relu', activation_output = 'softmax', 
                       dropout = 0.25):
    '''Model architecture inpsired from the paper below.
    
    References
    ----------
    Kim, Myungjong & Cao, Beiming & An, Kwanghoon & Wang, Jun. (2018). Dysarthric Speech Recognition Using Convolutional LSTM Neural Network. 10.21437/interspeech.2018-2250. 
    '''
    cnn = Sequential()
    cnn.add(Conv2D(feature_map_filters, kernel_size = kernel_size, activation = activation_layer))
    # non-overlapping pool_size 
    cnn.add(MaxPooling2D(pool_size = pool_size))
    cnn.add(Dropout(dropout))
    cnn.add(Flatten())

    # prepare stacked LSTM
    model = Sequential()
    model.add(TimeDistributed(cnn,input_shape = input_shape))
    model.add(LSTM(lstm_cells, return_sequences = True))
    model.add(LSTM(lstm_cells, return_sequences = True))
    model.add(Flatten())
    model.add(Dense(num_labels,activation = activation_output)) 
    settings = dict(input_shape = input_shape,
                    num_labels = num_labels,
                    kernel_size = kernel_size,
                    pool_size = pool_size,
                    activation_layer = activation_layer,
                    activation_output = activation_output,
                    lstm_cells = lstm_cells,
                    feature_map_filters = feature_map_filters,
                    dense_hidden_units = dense_hidden_units,
                    dropout = dropout)
    return model, settings
